import re
import requests
from fairNLP import Regex as Re
# Log = Log("FAIR.URL")

"""
    -> URL HANDLING
-> One period vs Two Periods?
    - The idea is all URLs have either 1 or 2 periods in the string.
        1. "verge.com"
        2. "www.verge.com"
        - Outside of this, there are only slashes. No more periods.
    - We figure out which it is in the current url
        - if 1 period, sitename is before last period.
        - if 2 periods, sitename is inbetween two periods.
    - We then parse the url to find the base site name.
"""
def get_site_name(url):
    """ -> PUBLIC -> Extract Base Site Name from URL <- """
    if verifyTwoPeriods(url):
        return extract_siteName_two_periods(url)
    return extract_siteName_one_period(url)

# -> get_site_name HELPER for verifying URL has one or two periods.
def verifyTwoPeriods(url: str):
    """ PRIVATE """
    count = 0
    for char in url:
        if char == '.':
            count += 1
    if count >= 2:
        return True
    return False

# -> get_site_name HELPER for urls with one period.
def extract_siteName_one_period(url):
    """ PRIVATE """
    i = 0
    slash_count = 0
    removal_index = 0
    for char in url:
        if char == "/":
            slash_count += 1
            if slash_count == 2:
                removal_index = i+1
        if char == '.':
            return url[removal_index:i]
        i += 1
    return url

# -> get_site_name HELPER for urls with two periods.
def extract_siteName_two_periods(url):
    """ PRIVATE """
    temp = ""
    start = 0
    end = 0
    i = 0
    periodCount = 0
    lastChar = ''
    # -> Part 1
    for char in url:
        if char == '/' and lastChar == '/':
            start = i+1
        if char == '.':
            periodCount += 1
            if periodCount == 2:
                end = i
        if end > 0:
            temp = url[start:end]
            break
        lastChar = char
        i += 1
    # -> Part 2
    n = 0
    match = ""
    for c in temp:
        if c == '.':
            match = temp[n+1:]
        n += 1
    return match

def extract_data_src_url(content: str):
    """ EXPERIMENTAL (under development still) """
    new = content.split("https://")
    for item in new:
        test = Re.contains(":", item)
        if test:
            continue
        if len(item) < 5:
            continue
        return "https://" + item
    return False

def is_valid_url(url):
    """ PUBLIC """
    try:
        response = requests.get(url)
        if response:
            print("URL is valid and exists on the internet")
            return True
    except requests.ConnectionError as e:
        print(f"URL does not exist on Internet, error=[ {e} ]")
        return False

def is_url(content: str):
    """ PUBLIC """
    try:
        if type(content) in [list, tuple]:
            for itemContent in content:
                match = re.search(r'http.?://.*/', add_http(itemContent))
                if match is not None:
                    return True, match.string
            return False, False
        else:
            match = re.search(r'http.?://.*/', add_http(content))
            return True, match.string if match is not None else False
    except Exception as e:
        print(f"Failed to regex findall. {content}, error=[ {e} ]")
        return False, False

def add_http(content):
    if content.startswith("//") or not content.startswith("http"):
        url1 = "https:" + content
        return url1
    return content

def avoid_url(url, avoid_list):
    """ EXPERIMENTAL """
    one = url.replace("https", "")
    two = one.replace("www", "")
    three = two.replace(".", "/")
    final = three.split("/")
    for avoid in avoid_list:
        if avoid in final:
            return True
    return False


import re
from FList import LIST
from FDate import DATE
# Log = Log("FAIR.Regex")

def remove_special_characters(text):
    """ DEPRECATED """
    newText = re.sub('[^a-zA-Z0-9]', ' ', text)
    return newText

def search(search_term, content):
    match = re.findall(fr'.*{search_term}.*', content)
    return match if len(match) >= 1 and match is not None else False

def contains(search_term, content):
    try:
        if type(content) in [list, tuple]:
            for itemContent in content:
                match = re.findall(fr'.*{search_term}.*', itemContent)
                if len(match) >= 1 and match is not None:
                    return True
            return False
        else:
            match = re.findall(fr'.*{search_term}.*', content)
            return True if len(match) >= 1 and match is not None else False
    except Exception as e:
        print(f"Failed to regex findall. {search_term}, error=[ {e} ]")
        return False

def contains_any(search_terms, content):
    search_terms = LIST.flatten(search_terms)
    for term in search_terms:
        temp = contains(term, content)
        if temp:
            return True
    return False

def str_contains_year(content):
    match = re.findall(r'\b[1-2][0-9][0-9][0-9]\b', content)
    if match and len(match) >= 1:
        if LIST.get(0, match) == "":
            return False

        newMatch = []
        for year in match:
            if len(year) > 4:
                continue
            if int(year) > int(DATE.get_current_year()):
                continue
            if int(year) < 1900:
                continue

            newMatch.append(year)
        return newMatch

def locate_term_in_str(term, content):
    match = re.search(fr'\b{term}\b', content)
    if match:
        return match
    return False


"""
1. March 02 2022
2. 03/02/2022
3. 03-02-2022
4. 2022-03-02
"""
def extract_date(content):
    if type(content) not in [str]:
        content = str(content)
    all = []
    months = DATE.months.keys()
    for item in months:
        all.append(item)
    month_variants = DATE.months.values()
    for item in month_variants:
        all.append(item)
    temp_all = LIST.flatten(all)
    _all = LIST.remove_duplicates(temp_all)
    for month_variant in _all:
        # -> Check for any Month or Month Variant in Full String
        locate_month_variant = locate_term_in_str(month_variant, content)
        if locate_month_variant:
            start = locate_month_variant.start()
            end = start + 15
            temp = content[start:end]
            # -> Check for Year immediately after Month
            find_year = str_contains_year(temp)
            year = LIST.get(0, find_year)
            locate_year = locate_term_in_str(year, temp)
            if locate_year:
                year_end = locate_year.end()
                temp2 = content[start:start+year_end]
                # -> Clean Result.
                filter = remove_special_characters(temp2)
                result = filter.replace("  ", " ").strip()
                return result
    return False


import FList.LIST
from FList import LIST
from nltk import WordNetLemmatizer
# Log = Log("FAIR.Language")

lemmatizer = WordNetLemmatizer()

"""
    -> Tokenizing/Splitting Words from a String.
"""

def to_words_v1(content: str):
    content = replace(content, ".", ",", ";", "\n", "  ")
    s = content.split(" ")
    newS = remove_empty_strings(s)
    return newS

def remove_empty_strings(list_of_strs: []):
    newS = []
    for word in list_of_strs:
        if word == '':
            continue
        newS.append(word)
    return newS

def replace(content, *args):
    for arg in args:
        content = content.replace(arg, " ")
    return content

def score_complete_tokenization(tokenization: dict):
    result = {}
    for key in tokenization.keys():
        token_list = tokenization[key]
        result[key] = score_words(token_list)
    return result

# @Ext.safe_args
def complete_tokenization_v2(content, toList=True):
    """ PUBLIC """
    content = FList.LIST.flatten(content)
    toStr = LIST.to_str(content)
    tokens = to_words_v1(toStr)
    bi_grams = to_x_grams(tokens, 2)
    tri_grams = to_x_grams(tokens, 3)
    quad_grams = to_x_grams(tokens, 4)
    if toList:
        return tokens + bi_grams + tri_grams + quad_grams
    else:
        return {"tokens": tokens, "bi_grams": bi_grams, "tri_grams": tri_grams, "quad_grams": quad_grams}

def to_x_grams(tokens, x):
    """ PUBLIC """
    if type(tokens) == str:
        tokens = to_words_v1(tokens)
    i = 0
    x_grams = []
    if len(tokens) < x:
        print("found none", tokens)
        return x_grams
    for _ in tokens:
        if i+x > len(tokens):
            break
        phrase = ""
        for c in range(x):
            phrase = combine_words(phrase, tokens[i+c])
        x_grams.append(phrase)
        i += 1
    return x_grams

def to_bi_grams_v2(tokens):
    """ PUBLIC HELPER """
    return to_x_grams(tokens, 2)

def to_tri_grams_v2(tokens):
    """ PUBLIC HELPER """
    return to_x_grams(tokens, 3)

def to_quad_grams_v2(tokens):
    """ PUBLIC HELPER """
    return to_x_grams(tokens, 4)

"""
    -> WORD EXPANSION 
"""
def expand_word(word: str) -> []:
    """ PUBLIC -> FAIR Expansion <- """
    word_lower = word.lower()
    word_upper = word.upper()
    word_first_capital = word[0].upper() + word[1:]
    word_stem = lemmatize_word(word)
    word_stem_first_capital = word_stem[0].upper() + word_stem[1:]
    return [word, word_lower, word_upper, word_first_capital, word_stem, word_stem_first_capital]

def lemmatize_word(word):
    """ PUBLIC """
    temp = remove_ing(word)
    if temp:
        return temp
    return lemmatizer.lemmatize(word)

"""
    -> FAIR UTILS
"""

def score_words(words):
    result = {}
    for word in words:
        if word in result.keys():
            tempValue = result[word]
            result[word] = tempValue + 1
        else:
            result[word] = 1
    return result

def combine_words(*words):
    """ Combines two strings together. """
    temp_word = ""
    if len(words) > 0:
        for word in words:
            temp_word += " " + word.strip()
        return temp_word.strip()
    return str(words).strip()

# @Ext.safe_args
def combine_args_str(content: str) -> str:
    temp = ""
    for item in content:
        temp += " " + str(item)
        print(temp)
    return str(temp).strip()

def remove_ing(word):
    if word.endswith("ing"):
        return word[:-3]
    elif word.endswith("ings"):
        return word[:-4]
    return False

def remove_apos(word):
    word = word.replace("'", "")
    return word



from FSON import DICT
from FDate import DATE
import uuid

# Log = Log("FAIR.Parser.HookupParser")

potential_keys = {
        "author": ["author", "authors", "screen_name", "user"],
        "title": ["title"],
        "body": ["body", "text", "full_text"],
        "description": ["description", "meta_description"],
        "published_date": ["published_date", "publish_date", "created_at", "parsely-pub-date"],
        "url": ["url", "urls"],
        "imgUrl": ["imgUrl", "meta_img", "profile_image_url"],
        "tags": ["tags", "parsely-tags", "keywords", "meta_keywords", "news_keywords", "parsely-section", "parsely-type"],
        "source_url": ["source_url", "source"],
        "tickers": ["tickers"],
        "companies": ["companies"],
    }


def keys(key) -> []:
    return potential_keys[key]

class Parser:

    @staticmethod
    def parse_list(data, parseAll=False):
        temp_final = []
        for item in data:
            hookup = Parser.parse(item, parseAll=parseAll)
            temp_final.append(hookup)
        return temp_final

    @staticmethod
    def parse(data, parseAll=False):
        print("Parsing data IN=[ {data} ]")
        try:
            hookup = Hookup()
            hookup.author = DICT.get_all_keys(data, keys("author"), force_type=True)
            hookup.title = DICT.get_all_keys(data, keys("title"), force_type=True)
            hookup.description = DICT.get_all_keys(data, keys("description"), force_type=True)
            hookup.body = DICT.get_all_keys(data, keys("body"), force_type=True)
            hookup.url = DICT.get_all_keys(data, keys("url"), force_type=True)
            hookup.img_url = DICT.get_all_keys(data, keys("imgUrl"), force_type=True)
            hookup.source = DICT.get_all_keys(data, keys("source_url"), force_type=True)
            hookup.tickers = DICT.get_all_keys(data, keys("tickers"))
            hookup.tags = DICT.get_all_keys(data, keys("tags"))
            temp_date = DICT.get_all_keys(data, keys("published_date"))
            hookup.published_date = DATE.parse_obj_to_month_day_year_str(temp_date)

            if parseAll:
                hookup.summary = DICT.get("summary", data)
                hookup.comments = DICT.get("comments", data)
                hookup.source_rank = DICT.get("source_rank", data)
                hookup.category = DICT.get("category", data)
                hookup.sentiment = DICT.get("sentiment", data)
                hookup.category_scores = DICT.get("category_scores", data)
                hookup.score = DICT.get("score", data)
                hookup.title_score = DICT.get("title_score", data)
                hookup.description_score = DICT.get("description_score", data)
                hookup.body_score = DICT.get("body_score", data)
                hookup.rank = DICT.get("rank", data)

            print(f"Parsing data OUT=[ {hookup} ]")
            return hookup
        except Exception as e:
            print(f"Failed to parse data=[ {data} ] error=[ {e} ]")
            return None

    @staticmethod
    def to_json(hookup):
        print(f"to_json: IN: {hookup}")
        json = {
            "id": hookup.id,
            "author": hookup.author,
            "title": hookup.title,
            "description": hookup.description,
            "body": hookup.body,
            "summary": hookup.summary,
            "tickers": hookup.tickers,
            "comments": hookup.comments,
            "published_date": hookup.published_date,
            "img_url": hookup.img_url,
            "url": hookup.url,
            "source": hookup.source,
            "source_rank": hookup.source_rank,
            "category": hookup.category,
            "sentiment": hookup.sentiment,
            "category_scores": hookup.category_scores,
            "score": hookup.score,
            "title_score": hookup.title_score,
            "description_score": hookup.description_score,
            "body_score": hookup.body_score,
            "rank": hookup.rank
        }
        print(f"to_json: OUT: {json}")
        return json


class Hookup(object):
    id = ""
    processed = False
    author = ""
    # -> Content
    title: str = ""
    description = ""
    body = ""
    comments = ""
    summary = ""
    tags = ""
    # -> Date
    published_date = ""
    # -> Urls
    img_url = ""
    url = ""
    # -> Source
    source = ""
    source_rank = 0
    category = ""
    # -> Process...
    tickers = {}  # { "stocks": { "BB": 22 }, "crypto": { "MANA": 16 } }
    sentiment = {}
    category_scores = {}
    score = 0
    title_score = 0
    description_score = 0
    body_score = 0
    rank = 0

    def __init__(self):
        self.id = str(uuid.uuid4())

    @staticmethod
    def convert_list_to_archive_json(list_of_hookups):
        temp = []
        if not list_of_hookups:
            return temp
        for hookup in list_of_hookups:
            j = Hookup.to_archive_json(hookup)
            temp.append(j)
        return temp

    @staticmethod
    def to_archive_json(data):
        print(f"to_archive_json: IN: {data}")
        json = {
            "author": DICT.get("author", data),
            "title": DICT.get("title", data),
            "description": DICT.get("description", data),
            "body": DICT.get("body", data),
            "tags": DICT.get("tags", data),
            "published_date": DICT.get("published_date", data),
            "img_url": DICT.get("img_url", data),
            "url": DICT.get("url", data),
            "source": DICT.get("source", data),
        }
        print(f"to_archive_json: OUT: {json}")
        return json

    @staticmethod
    def to_file_json(data):
        print(f"to_file_json: IN: {data}")
        try:
            json = {
                "source": DICT.get("source", data),
                "source_rank": DICT.get("source_rank", data),
                "author": DICT.get("author", data),
                "category": DICT.get("category", data),
                "description": DICT.get("description", data),
                "url": DICT.get("url", data),
                "tickers": DICT.get("tickers", data),
                "sentiment": DICT.get("sentiment", data),
                "title_score": DICT.get("title_score", data),
                "description_score": DICT.get("description_score", data),
                "body_score": DICT.get("body_score", data),
                "score": DICT.get("score", data),
                "rank": DICT.get("rank", data),
                "category_scores": DICT.get("category_scores", data)
            }
            return json
        except Exception as e:
            print(f"Failed to parse hookup into json. error=[ {e} ]")
            return data

    @staticmethod
    def sort_by_date(list_of_hookups) -> dict:
        try:
            by_date = {}
            for hookup in list_of_hookups:
                temp_date = DICT.get("published_date", hookup)
                if by_date.__contains__(temp_date):
                    temp_list = by_date[temp_date]
                    temp_list.append(hookup)
                    by_date[temp_date] = temp_list
                else:
                    by_date[temp_date] = [hookup]
            return by_date
        except Exception as e:
            print(f"Failed to sort hookups by date. error=[ {e} ]")
            return list_of_hookups

    @staticmethod
    def get_top_ranked(hookups):
        highest = []
        if len(hookups) < 1 or hookups is None:
            return highest
        highest = hookups[0]
        for hookup in hookups:
            if hookup.rank > highest.rank:
                highest = hookup
        return highest

    @staticmethod
    def sort_hookups_by_rank(hookups, reversed=True):
        # print(f"sort_hookups_by_rank: IN: {hookups}")
        sorted_hookups = sorted(hookups, key=lambda k: k.get("rank"), reverse=reversed)
        return sorted_hookups

    @staticmethod
    def sort_hookups_by_score(hookups, reversed=True):
        # print(f"sort_hookups_by_score: IN: {hookups}")
        sorted_hookups = sorted(hookups, key=lambda k: k.get("score"), reverse=reversed)
        return sorted_hookups



potential_keys = {
        "author": ["author", "authors", "screen_name", "user"],
        "title": ["title"],
        "body": ["body", "text", "full_text"],
        "description": ["description", "meta_description"],
        "published_date": ["published_date", "publish_date", "created_at", "parsely-pub-date"],
        "url": ["url", "urls"],
        "imgUrl": ["imgUrl", "meta_img", "profile_image_url"],
        "tags": ["tags", "parsely-tags", "keywords", "meta_keywords", "news_keywords", "parsely-section", "parsely-type"],
        "source_url": ["source_url", "source"],
        "tickers": ["tickers"],
        "companies": ["companies"],
    }

def keys(key) -> []:
    return potential_keys[key]


import os

MASTER_PATH = os.getcwd()

ENGLISH = "ENGLISH"
GERMAN = "GERMAN"
LANGUAGE = ENGLISH

ERROR = 0  # -> Show ERROR only
INFO = 1  # -> Show ERROR and INFO
DEBUG = 2  # -> Show ERROR, INFO and DEBUG
VERBOSE = 3  # -> Show ERROR, INFO, DEBUG AND VERBOSE
LOG_LEVEL = DEBUG

LATEST = 0  # 0 == LATEST
NLP_VERSION = 1

removal_words = ["in", "the", "by", "to", "of", "as", "on",
                 "is", "now", "be", "will", "a", "it", "it's",
                 "its", "at", "into", "for", "that", "you",
                 "and", "or", "new", "are", "a"]

table = {
         ord('ä'): 'ae',
         ord('ö'): 'oe',
         ord('ü'): 'ue',
         ord('ß'): 'ss',
       }


from fairConfig import fairFig
from pathlib import Path
from FSON import DICT
import json
import os

# Log = Log("FAIR.Cache.LocalCache")

cache_path = fairFig.MASTER_PATH + "/Data/Cache"

class LocalCache:
    processName = ""
    file = None
    data = None

    def __init__(self, processName, clearCache=False):
        self.processName = processName
        # if clearCache:
        #     self.clear_cache()
        # self.file = open(f"{cache_path}/{self.processName}.json", "w+")
        temp = Path(f"{cache_path}/{self.processName}.json")
        temp.touch(exist_ok=True)
        self.file = open(temp)
        print(f"Caching Enabled for: {processName}.")

    def data_to_json(self, data):
        temp = []
        for item in data:
            temp.append(item)
        return temp

    def save_data(self, data):
        if type(data) is list:
            data = { "data": data }
        data = self.data_to_json(data)
        data = {"data": data}
        try:
            with open(f'{cache_path}/{self.processName}.json', 'w') as f:
                json.dump(data, f, sort_keys=True, indent=4)
            print(f"Saved File {self.processName}.json to Cache Directory")
        except Exception as e:
            print(f"Error saving dict'd dict. error=[ {e} ]")
            return None

    def load_data(self):
        try:
            file = open(f"{cache_path}/{self.processName}.json")
            temp = json.load(file)
            self.data = DICT.get("data", temp)
            return self.data
        except Exception as e:
            print(f"No File Found. error=[ {e} ]")
            return None

    def clear_cache(self):
        try:
            os.remove(f"{cache_path}/{self.processName}.json")
            print(f"Cleared cache file {self.processName}.json.")
        except Exception as e:
            print(f"Failed to remove cache file. error=[ {e} ]")



from FSON import DICT
from FDate import DATE
from fairParser import Keys

# Log = Log("FAIR.Parser.JsonParser")

keys = Keys.keys

def parse(data, parseAll=False) -> {}:
    print(f"Parsing data IN=[ {data} ]")
    try:
        json_obj = {}
        json_obj["author"] = DICT.get_all_keys(data, keys("author"), force_type=True)
        json_obj["title"] = DICT.get_all_keys(data, keys("title"), force_type=True)
        json_obj["description"] = DICT.get_all_keys(data, keys("description"), force_type=True)
        json_obj["body"] = DICT.get_all_keys(data, keys("body"), force_type=True)
        json_obj["url"] = DICT.get_all_keys(data, keys("url"), force_type=True)
        json_obj["img_url"] = DICT.get_all_keys(data, keys("imgUrl"), force_type=True)
        json_obj["source"] = DICT.get_all_keys(data, keys("source_url"), force_type=True)
        json_obj["tickers"] = DICT.get_all_keys(data, keys("tickers"))
        json_obj["tags"] = DICT.get_all_keys(data, keys("tags"))
        temp_date = DICT.get_all_keys(data, keys("published_date"))
        json_obj["published_date"] = DATE.parse_obj_to_month_day_year_str(temp_date)

        if parseAll:
            json_obj["summary"] = DICT.get("summary", data)
            json_obj["comments"] = DICT.get("comments", data)
            json_obj["source_rank"] = DICT.get("source_rank", data)
            json_obj["category"] = DICT.get("category", data)
            json_obj["sentiment"] = DICT.get("sentiment", data)
            json_obj["category_scores"] = DICT.get("category_scores", data)
            json_obj["score"] = DICT.get("score", data)
            json_obj["title_score"] = DICT.get("title_score", data)
            json_obj["description_score"] = DICT.get("description_score", data)
            json_obj["body_score"] = DICT.get("body_score", data)

        print(f"Parsing data OUT=[ {json_obj} ]")
        return json_obj
    except Exception as e:
        print(f"Failed to parse data=[ {data} ], error=[ {e} ]")
        return None


import re

# from Config import tif_tickers
# from Config.Topic import Topic
# Log = Log("FAIR.Ticker")
"""
    -> TICKER EXTRACTION
"""
# def extract_classify_tickers_v2(content):
#     """ PUBLIC """
#     potential_tickers = find_tickers_strict_v2(content)
#     return classify_tickers(potential_tickers)

def find_tickers_strict_v2(content):
    """ PRIVATE ->
                Pretty Damn Accurate
    -> MUST BE all capitalized or will not get found.
    -> "xrp": False "XRP": True
    """
    match = re.findall(r'[^\s$:(][A-Z]{1,5}(?=\s|\)|/|,|\.)', content)
    return match if match is not None else False

def find_tickers_light_v2(content):
    """ -> PRIVATE -> Allows LowerCase, drastically decreasing the accuracy. <- """
    match = re.findall(r'[^\s$:(][A-Za-z]{1,5}(?=\s|\)|/|,|\.)', content)
    return match if match is not None else False

# def classify_tickers(list_of_potential_tickers):
#     """ -> PRIVATE -> Helper Method for classify_ticker() <- """
#     Log.v(f"classify_tickers IN: {list_of_potential_tickers}")
#     results = []
#     stop_words = Topic.load_stopwords()
#     for word in list_of_potential_tickers:
#         temp = classify_ticker_v2(word, stopWords=stop_words)
#         if temp:
#             results.append(temp)
#     Log.v(f"classify_tickers OUT: {results}")
#     return results

# def classify_ticker_v2(potential_ticker, stopWords=None):
#     """ -> PRIVATE -> Tries to figure out if ticker is a Stock or Crypto <- """
#     stock = "Stock"
#     crypto = "Crypto"
#     word = potential_ticker.upper()
#     if stopWords:
#         if word in stopWords or word.lower() in stopWords:
#             return False
#     # -> BlackList
#     if word in tif_tickers.blacklist or word.lower() in tif_tickers.blacklist:
#         return False
#     # -> Stock -> if in stock tickers and not in crypto tickers
#     if word in tif_tickers.STOCK_TICKERS:
#         if word not in tif_tickers.CRYPTO_TICKERS:
#             Log.v(f"classify_tickers: Word={word}, Outcome={(stock, word)}")
#             return stock, word
#     # -> CRYPTO -> if word is in Special Tickers, crypto.
#     if word in tif_tickers.SPECIAL_TICKERS.keys():
#         Log.v(f"classify_tickers: Word={word}, Outcome={(crypto, word)}")
#         return crypto, word
#     # -> CRYPTO -> if word is in crypto tickers
#     if word in tif_tickers.CRYPTO_TICKERS:
#         Log.v(f"classify_tickers: Word={word}, Outcome={(crypto, word)}")
#         return crypto, word
